# -*- coding: utf-8 -*-

import sys
from collections import defaultdict
import math
import re
import random
import operator
import pickle


def read_data( filename ):
    """ reads tagged sentences from a file """
    sentences = []; words = []; tags = []
    with open(filename) as file:
        for line in file:
            line = line.rstrip()
            if line == '':
                # empty line indicates end of sentence
                words = ['<s>'] + words + ['<s>']
                tags = ['<s>'] + tags + ['<s>']
                sentences.append((words, tags))
                words = []; tags = []
            else:
                word, tag = line.split("\t")
                words.append(word)
                tags.append(tag)
    return sentences


def log_add(a, b):
    return a + math.log1p(math.exp(b - a)) if a >= b else log_add(b,a)


def get_case( word ):
    if word[0].isupper():
        return 'U'
    if word[0].islower():
        return 'L'
    return 'O'


class CRFTagger:

    def __init__(self, train_filename, learning_rate ):

        self.learning_rate = learning_rate
        self.train_sentences = read_data( train_filename )
        self.tagset = set([tag for (_, tags) in self.train_sentences for tag in tags])
        self.tagset.remove('<s>')
        self.weight = {}
        self.last_update = {}
        self.iteration = 0
        self.training = True
        self.num_iterations = 100000
        self.max_substr_len = 6
        self.num_best_tags1 = 8
        self.num_best_tags2 = 4

        self.shape = {}
        self.substrings = {}

    def init_caches( self ):
        """ initializes the dictionaries used for caching """
        self.lex_features = {}
        self.lex_score = {}
        self.context_features = {}
        self.context_score = {}
        self.forward = None
        self.backward = None


    def get_shape( self, word ):
        """ computes the shape of a word with caching """
        if word not in self.shape:
            s = re.sub(r'[A-ZÄÖÜ]','X',word)
            s = re.sub(r'[a-zäöüß]','x',s)
            s = re.sub(r'[0-9]','9',s)
            s = re.sub(r'(.)\1+',r'\1',s)
            self.shape[word] = s
        return self.shape[word]

    def get_substrings( self, word ):
        if word not in self.substrings:
            w = '#' + word + '#'
            S = []
            L = len(w)
            for l in range(2, self.max_substr_len+1):
                for i in range(L):
                    if i+l > L:
                        break
                    S.append( w[i:i+l] )
            self.substrings[word] = S
        return self.substrings[word]

    def get_lex_features(self, tag, pos):
        """ computation of the list of lexical features """
        if (tag,pos) not in self.lex_features:
            curr_word = self.word[pos]
            lc_word   = curr_word.lower()
            prev_word = self.word[pos-1]
            next_word = self.word[pos+1] if pos<len(self.word)-1 else '<s>'

            features = ['TW_'+ tag + '_' + curr_word,
                        'TL_'+ tag + '_' + lc_word,
                        'TF_'+ tag + '_' + self.get_shape(curr_word),
                        'TWW+1_' + tag + '_' + curr_word + '_' + next_word,
                        'TW-1W_' + tag + '_' + prev_word + '_' + curr_word,
                        'TC_' + tag + '_' + get_case(curr_word)]

            # substring features
            for s in self.get_substrings( lc_word ):
                features.append( 'TSUBS_' + tag + '_' + s )

            self.lex_features[(tag, pos)] = features
        return self.lex_features[(tag, pos)]


    def get_context_features(self, prevtag, tag, pos):
        """ computation of the list of features which depend on the previous tag """
        curr_word = self.word[pos]
        prev_word = self.word[pos-1]
        if (prevtag, tag, pos) not in self.context_features:
            features = ['TT_' + prevtag + '_' + tag,
                        'TTW_' + prevtag + '_' + tag + '_' + curr_word,
                        'TTW-1_' + prevtag + '_' + tag + '_' + prev_word]
            self.context_features[(prevtag, tag, pos)] = features
        return self.context_features[(prevtag, tag, pos)]


    def multiply_weights(self, features):
        """ multiplication of a feature vector and a weight vector """
        return sum([self.weight.get(f, 0.0) for f in features])


    def compute_lex_score(self, tag, pos):
        """ computes the product of the weight vector and the lexical feature vector """
        if (tag,pos) not in self.lex_score:
            features = self.get_lex_features(tag, pos)
            self.lex_score[(tag, pos)] = self.multiply_weights( features )
        return self.lex_score[(tag, pos)]

    def compute_context_score(self, prevtag, tag, pos):
        """ computes the product of the weight vector and the context feature vector """
        if (prevtag,tag,pos) not in self.context_score:
            features = self.get_context_features(prevtag, tag, pos)
            self.context_score[(prevtag, tag, pos)] = self.multiply_weights( features )
        return self.context_score[(prevtag, tag, pos)]

    def get_tags( self, i ):
        """ computes the best tags of a word based on the lexical features """
        if i == len(self.word) - 1:
            return ['<s>']
        score = { tag: self.compute_lex_score(tag, i) for tag in self.tagset }
        return sorted(score, key=score.get, reverse=True)[:self.num_best_tags1]
    
    def compute_forward_scores( self ):
        self.forward = [{'<s>':0}]
        for i in range(1, len(self.word)):
            self.forward.append(dict())
            for tag2 in self.get_tags(i):
                lexscore = self.compute_lex_score(tag2, i)
                for tag1 in self.forward[i-1]:
                    contextscore = self.compute_context_score(tag1,tag2,i)
                    score = self.forward[i-1][tag1] + lexscore + contextscore
                    if tag2 in self.forward[i]:
                        self.forward[i][tag2] = log_add(self.forward[i][tag2], score)
                    else:
                        self.forward[i][tag2] = score
            # get the sorted list of key-value pairs
            tmp = sorted(self.forward[i].items(), key=operator.itemgetter(1), reverse=True)
            # keep the best ones
            self.forward[i] = dict(tmp[:self.num_best_tags2])

    def compute_backward_scores( self ):
        self.backward = [{} for _ in self.word]
        self.backward[-1] = {'<s>':0}
        for i in range(len(self.word)-1, 0, -1):
            for tag2 in self.forward[i]:
                lexscore = self.compute_lex_score(tag2, i)
                for tag1 in self.forward[i-1]:
                    contextscore = self.compute_context_score(tag1,tag2,i)
                    score = lexscore + contextscore + self.backward[i][tag2]
                    if tag1 in self.backward[i-1]:
                        self.backward[i-1][tag1] = log_add(self.backward[i-1][tag1], score)
                    else:
                        self.backward[i-1][tag1] = score

    def expected_feature_counts( self ):
        self.init_caches()
        self.compute_forward_scores()
        self.compute_backward_scores()

        count = defaultdict(float)
        for i in range(1, len(self.word)):
            for tag2 in self.forward[i]:
                p = math.exp(self.forward[i][tag2] + self.backward[i][tag2] - self.forward[-1]['<s>'])
                for f in self.get_lex_features(tag2, i):
                    count[f] += p
                lexscore = self.compute_lex_score(tag2, i)
                for tag1 in self.forward[i-1]:
                    contextscore = self.compute_context_score(tag1,tag2,i)
                    p = math.exp(self.forward[i-1][tag1] + lexscore + contextscore + self.backward[i][tag2] - self.forward[-1]['<s>'])
                    for f in self.get_context_features(tag1, tag2, i):
                        count[f] += p
        return count


    def observed_feature_counts( self ):
        count = defaultdict(float)
        for i in range(1, len(self.word)):
            for f in self.get_lex_features(self.tag[i], i):
                count[f] += 1
            for f in self.get_context_features(self.tag[i-1], self.tag[i], i):
                count[f] += 1
        return count


    def save_parameters( self, filename ):
        with open(filename,"wb") as file:
            pickle.dump( self.tagset, file )
            pickle.dump( self.weight, file )

    def update_weights( self, feature_counts, sign ):
        lr = sign * self.learning_rate
        for f,v in feature_counts.items():
            if f in self.weight:
                self.weight[f] += lr * v
            else:
                self.weight[f] = lr * v
                self.last_update[f] = self.iteration
                

    def train( self ):
        for self.iteration in range(self.num_iterations):
            sys.stderr.write('\r'+str(self.iteration)+'  ')
            self.word, self.tag = random.choice(self.train_sentences)

            exp_counts = self.expected_feature_counts()
            obs_counts = self.observed_feature_counts()
            self.update_weights( obs_counts, +1 )
            self.update_weights( exp_counts, -1 )
